/*
 bitmap.cpp - Bitmap Library

 Copyright (c) 2023 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#pragma once

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/stat.h>

#define BITMAPFILETYPE "BM"

class Bitmap
{
  protected:

    typedef struct __attribute__((packed))
    {
      uint8_t rgbBlue;
      uint8_t rgbGreen;
      uint8_t rgbRed;
      uint8_t rgbReserved;
    } RGBQUAD;

    typedef struct __attribute__((packed))
    {
      //
      // BITMAPFILEHEADER
      //
      uint16_t bfType;
      uint32_t bfSize;
      uint16_t bfReserved1;
      uint16_t bfReserved2;
      uint32_t bfOffBits;
      //
      // BITMAPINFOHEADER
      //
      uint32_t biSize;
      int32_t  biWidth;
      int32_t  biHeight;
      uint16_t biPlanes;
      uint16_t biBitCount;
      uint32_t biCompression;
      uint32_t biSizeImage;
      int32_t  biXPelsPerMeter;
      int32_t  biYPelsPerMeter;
      uint32_t biClrUsed;
      uint32_t biClrImportant;
      RGBQUAD  biColors[0];
    } BITMAPHEADER;

    BITMAPHEADER *_bitmap;

    void bitmap(BITMAPHEADER *bh)
    {
      if (_bitmap)
        free(_bitmap);
      _bitmap = bh;
    }

    inline
    size_t lineBytes(int32_t width, uint16_t bitCount)
    {
      return ((abs(width) * (bitCount ? bitCount : 1) + 31) & ~31) >> 3;
    }

    inline
    size_t lineBytes(BITMAPHEADER *bh)
    {
      return lineBytes(bh->biWidth, bh->biBitCount);
    }

  public:

    Bitmap(void) : _bitmap(nullptr)
    {
    }

    Bitmap(Bitmap& from) : Bitmap()
    {
      copy(from);
    }

    Bitmap(int32_t width, int32_t height, uint16_t bitCount) : Bitmap()
    {
      create(width, height, bitCount);
    }

    ~Bitmap(void)
    {
      clear();
    }

    void clear(void)
    {
      bitmap(nullptr);
    }

    bool copy(Bitmap& from)
    {
      BITMAPHEADER *bh = nullptr;

      if (from._bitmap)
      {
        if ((bh = (BITMAPHEADER *)malloc(from._bitmap->bfSize)))
        {
          memcpy(bh, from._bitmap, from._bitmap->bfSize);
          bitmap(bh);
        }
      }
      return bh != nullptr;
    }

    int32_t width(void)
    {
      return _bitmap ? _bitmap->biWidth : 0;
    }

    int32_t height(void)
    {
      return _bitmap ? _bitmap->biHeight : 0;
    }

    uint32_t bitCount(void)
    {
      return _bitmap ? _bitmap->biBitCount : 0;
    }

    size_t size(void)
    {
      return _bitmap ? _bitmap->biSizeImage : 0;
    }

    size_t lineBytes(void)
    {
      return _bitmap ? lineBytes(_bitmap) : 0;
    }

    char *image(void)
    {
      return _bitmap ? (char *)_bitmap + _bitmap->bfOffBits : nullptr;
    }

    bool read(const char *file)
    {
      bool rv = false;
      BITMAPHEADER *bh;
      struct stat stb;
      FILE *fp;

      if ((fp = fopen(file, "rb")))
      {
        if (fstat(fileno(fp), &stb) != -1)
        {
          if (stb.st_size >= sizeof(BITMAPHEADER))
          {
            if ((bh = (BITMAPHEADER *)malloc(stb.st_size)))
            {
              if (fread(bh, 1, stb.st_size, fp) != stb.st_size)
              {
                free(bh);
                fprintf(stderr, "[Error] [%s] file read error\n", file);
              }
              else if (memcmp(&bh->bfType, BITMAPFILETYPE, sizeof(bh->bfType)) != 0)
              {
                free(bh);
                fprintf(stderr, "[Error] [%s] invalid BITMAPFILEHEADER.bfType\n", file);
              }
              else if (bh->bfOffBits < sizeof(BITMAPHEADER) + sizeof(RGBQUAD) * bh->biClrUsed)
              {
                free(bh);
                fprintf(stderr, "[Error] [%s] invalid BITMAPFILEHEADER.bfOffBits\n", file);
              }
              else if ((bh->biCompression == 0) && ((bh->bfOffBits + lineBytes(bh) * bh->biHeight > stb.st_size)))
              {
                free(bh);
                fprintf(stderr, "[Error] [%s] invalid BITMAPFILEHEADER.bfOffBits\n", file);
              }
              else
              {
                bh->bfSize   = stb.st_size;
                bh->biPlanes = 1;
                if (bh->biCompression == 0)
                  bh->biSizeImage = lineBytes(bh) * bh->biHeight;
                bitmap(bh);
                rv = true;
              }
            }
          }
        }
        fclose(fp);
      }
      return rv;
    }

    bool write(const char *file)
    {
      bool rv = false;
      FILE *fp;

      if (_bitmap)
      {
        if ((fp = fopen(file, "wb+")))
        {
          rv = fwrite(_bitmap, 1, _bitmap->bfSize, fp) == _bitmap->bfSize;
          fclose(fp);
        }
      }
      return rv;
    }

    void create(int32_t width, int32_t height, uint16_t bitCount)
    {
      BITMAPHEADER *bh;

      size_t hs = sizeof(BITMAPHEADER) + sizeof(RGBQUAD) * (bitCount ? 1 << bitCount : 0);
      size_t is = lineBytes(width, bitCount) * abs(height);
      size_t fs = hs + is;
      if ((bh = (BITMAPHEADER *)calloc(fs, 1)))
      {
        memcpy(&bh->bfType, BITMAPFILETYPE, sizeof(bh->bfType));
        bh->bfSize      = fs;
        bh->bfOffBits   = hs;
        bh->biWidth     = width;
        bh->biHeight    = height;
        bh->biPlanes    = 1;
        bh->biBitCount  = bitCount ? bitCount : 1;
        bh->biSizeImage = is;
        bh->biClrUsed   = bitCount ? 1 << bitCount : 0;
        bitmap(bh);
        color(bh->biClrUsed - 1, 0xFF, 0xFF, 0xFF);
      }
    }

    void color(uint8_t index, uint8_t blue, uint8_t green, uint8_t red)
    {
      if (_bitmap)
      {
        if (index < _bitmap->biClrUsed)
          _bitmap->biColors[index] = { .rgbBlue = blue, .rgbGreen = green, .rgbRed = red };
      }
    }
};
