/*
 ftdisio.h - Serial Library for Linux ftdi_sio driver

 Copyright (c) 2022 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#pragma once

#include <string>
#include <vector>
#include <stdlib.h>
#include <dirent.h>
#include "serial.h"

class FTDISIO : public Serial
{
  public:

    const char *DEVICE = "/dev/serial/by-id";

    FTDISIO(void) {}
    virtual ~FTDISIO(void) {}

    int getDeviceCount(void)
    {
      _portlist.clear();
      DIR *dirp = opendir(DEVICE);
      if (dirp)
      {
        struct dirent *dent;
        while ((dent = readdir(dirp)))
        {
          if (dent->d_type & DT_LNK)
          {
            std::vector<std::string> parts;
            port_t port;
            if (split(parts, dent->d_name, '-').size() == 4)
            {
              port.number = atoi(parts[3].substr(4).c_str());
              if (split(parts, parts[1], '_').size() == 3)
              {
                port.serial      = parts[2];
                port.description = parts[1];
                _portlist.push_back(port);
              }
            }
          }
        }
        closedir(dirp);
      }
      return _portlist.size();
    }

    int getComPortNumber(int index)
    {
      return (size_t)index < _portlist.size() ? _portlist[index].number : -1;
    }

    bool getComPortName(int index, char *buf)
    {
      strcpy(buf, "(null)");
      int num = getComPortNumber(index);
      if (num >= 0)
      {
        sprintf(buf, "ttyUSB%d", num);
        return true;
      }
      return false;
    }

    bool getDeviceSerial(int index, char *buf)
    {
      strcpy(buf, "(null)");
      if ((size_t)index < _portlist.size())
      {
        strcpy(buf, _portlist[index].serial.c_str());
        return true;
      }
      return false;
    }

    bool getDeviceDescription(int index, char *buf)
    {
      strcpy(buf, "(null)");
      if ((size_t)index < _portlist.size())
      {
        strcpy(buf, _portlist[index].description.c_str());
        return true;
      }
      return false;
    }

    int open(const char *device,
         unsigned long speed = 115200,
         SERIAL_FRAME frame = SERIAL_FRAME_8N1,
         SERIAL_FLOW flow = SERIAL_FLOW_NONE,
         unsigned long timeout = 1000) override
    {
      char *eos, name[64];
      int index = strtoul(device, &eos, 10);
      int count = getDeviceCount();
      if (*eos)
      {
        for (index = 0; index < count; ++index)
        {
          if (getDeviceSerial(index, name))
          {
            if (strcmp(device, name) == 0)
            {
              if (getComPortName(index, name))
                device = name;
              break;
            }
          }
        }
      }
      else if ((index >= 0) && (index < count))
      {
        if (getComPortName(index, name))
          device = name;
      }
      return Serial::open(device, speed, frame, flow, timeout);
    }

  private:

    std::vector<std::string>& split(std::vector<std::string>& parts, const std::string str, char c)
    {
      parts.clear();
      for (size_t pos = 0; pos != std::string::npos; )
      {
        size_t found = str.find_first_of(c, pos);
        if (found == std::string::npos)
          parts.push_back(str.substr(pos));
        else
          parts.push_back(str.substr(pos, found++ - pos));
        pos = found;
      }
      return parts;
    }

    typedef struct
    {
      int number;
      std::string serial;
      std::string description;
    } port_t;

    std::vector<port_t> _portlist;
};
