/*
 nameplate.cpp - Name Plate Manager

 Copyright (c) 2023 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <signal.h>
#include <time.h>
#include "ftdisio.h"
#include "bridge.h"
#include "hostdef.h"
#include "bitmap.h"

#define IMAGE_POLLING_TIME 10000   // ms
#define IMAGE_FILENAME     "nameplate.bmp"
#define IMAGE_WIDTH        800
#define IMAGE_HEIGHT       480
#define IMAGE_BITCOUNT     1
#define IMAGE_BYTES(w, h)  ((((w) + 7) / 8) * (h))

#define SERIAL_SPEED       1000000 // 1Mbps
#define SERIAL_TIMEOUT     500    // ms
#define SERIAL_RETRY_COUNT 10

#define EPAPER_FAST_IO     1
#define DEBUG_RETRY        1

#define SECONDS_OF_DAY     (24 * 60 * 60)

#define LOGFILE "nameplate.log"
#define ERRFILE "nameplate.err"

int interval = SECONDS_OF_DAY;
int origin   = 0;

char *prog_path;
char *prog_name;
int   exitcode;
bool  debug;
const char *plates[EPD_COLOR_MAX] = { nullptr, IMAGE_FILENAME, IMAGE_FILENAME ".red" };
Bitmap     *images[EPD_COLOR_MAX] = { nullptr, new Bitmap()  , new Bitmap()          };
bool    update;
FTDISIO serial;
Bridge::packet_t request;
Bridge::packet_t response;

int width    = IMAGE_WIDTH;
int height   = IMAGE_HEIGHT;
int bitcount = IMAGE_BITCOUNT;

void writeFile(FILE *fp, const char *path, const char *mode, bool timestamp, const char *format, ...)
{
  va_list args;

  va_start(args, format);
  if (fp)
    vfprintf(fp, format, args);
  if ((fp = fopen(path, mode)))
  {
    if (timestamp)
    {
      time_t t = time(NULL);
      struct tm stm = *localtime(&t);
      fprintf(fp, "%02d-%02d-%02d %02d:%02d:%02d ",
        stm.tm_year + 1900,
        stm.tm_mon  + 1,
        stm.tm_mday,
        stm.tm_hour,
        stm.tm_min,
        stm.tm_sec
      );
    }
    vfprintf(fp, format, args);
    fclose(fp);
  }
  va_end(args);
}

int readError(void)
{
  int rv = 0;
  FILE *fp;

  if ((fp = fopen(ERRFILE, "r")))
  {
    fscanf(fp, "%d", &rv);
    fclose(fp);
  }
  return rv;
}

time_t getLocalTime(void)
{
  time_t now = SECONDS_OF_DAY;
  struct tm stm = *localtime(&now);
  int tzoffs = (stm.tm_mday == 1 ? stm.tm_hour - 24 : stm.tm_hour);
  return time(NULL) + tzoffs * 60 * 60;
}

int calcWakeupTime(int device, int interval, int orgin = 0)
{
  static time_t last;
  static int    diff;
  static int    error;

  time_t now = getLocalTime();
  int off = (now - orgin) % interval;
  if (off >= (interval >> 1))
    off -= interval;
  if (diff != device)
    error = readError();
  else
  {
    int err = (now - last) - diff;
    if (abs(err) <= diff * 3 / 100) // err < 3% ?
    {
      // low pass filter
      error += (err - error) * 30 / 100;
      writeFile(NULL, ERRFILE, "w", false, "%d", error);
    }
  }
  last = now;
  return diff = interval - off - error;
}

void dump_packet(const Bridge::packet_t& packet, const char *prefix)
{
  if (debug)
  {
    const unsigned char *p = (const unsigned char *)&packet;
    if (prefix)
      fprintf(stdout, "%s", prefix);
    for (int i = 0; i <= packet.length; ++i)
      fprintf(stdout, "%02X ", *p++);
    fprintf(stdout, "\n");
  }
}

bool device_open(const char *name, long speed)
{
  if (serial.open(name, speed, SERIAL_FRAME_8N1, SERIAL_FLOW_NONE, SERIAL_TIMEOUT) < 0)
  {
    fprintf(stdout, "[ERROR] open(\"%s\"): %s\n", name, strerror(errno));
    return false;
  }
  return true;

}

void device_close(void)
{
  serial.close();
}

int device_read(Bridge::packet_t& packet, long timeout = -1)
{
  ssize_t n, cnt;
  for (cnt = packet.length = 0; cnt <= packet.length; cnt += n)
  {
    if ((n = serial.read((uint8_t *)&packet + cnt, packet.length - cnt + sizeof(packet.length), timeout)) <= 0)
      break;
  }
  if (cnt == packet.length + sizeof(packet.length))
  {
    //
    // dump packet
    //
    dump_packet(packet, "->");
    //
    // read packet
    //
    if (Bridge::csum((uint8_t *)&packet, cnt) == 0)
    {
      --packet.length; // remove csum
      return 1;
    }
  }
  return n < 0 ? -1 : 0;
}

bool device_write(Bridge::packet_t& packet)
{
  ssize_t rv, size = sizeof(packet.length) + packet.length;;
  //
  // calculate csum
  //
  *((uint8_t *)&packet + packet.length) = Bridge::csum((uint8_t *)&packet, packet.length);
  //
  // dump packet
  //
  dump_packet(packet, "<-");
  //
  // write packet
  //
  if ((rv = serial.write(&packet, size)) == -1)
    fprintf(stdout, "[ERROR] write(): %s\n", strerror(errno));
  else if (rv != size)
    fprintf(stdout, "[ERROR] write(): %d != %d\n", rv, size);
  else
    return true;
  return false;
}

bool device_command(Bridge::packet_t& packet, bool debugout = false)
{
  bool rv = false;
  int i;

  packet.type = 0;
  for (i = 0; i < SERIAL_RETRY_COUNT; ++i)
  {
    //
    // write packet
    //
    if (!device_write(packet))
      break;
    //
    // read packet
    //
    if (device_read(response) > 0)
    {
      rv = ((command_t *)response.data)->args.status == 0;
      break;
    }
    //
    // re-transmit
    //
    if (!(packet.type & Bridge::TYPE_RETRANSMIT))
    {
      packet.type   = Bridge::TYPE_RETRANSMIT;
      packet.length = offsetof(__typeof__(packet), data) + sizeof(packet.data[0]);
    }
    if (debugout)
    {
      fprintf(stdout, ".");
      fflush(stdout);
    }
  }
  return rv;
}

#define PACKET_LENGTH(f) ((size_t)(char *)&f - (size_t)(char *)&packet + sizeof(f))

int device_paint(Bridge::packet_t& packet, int color, Bitmap *bitmap, bool debugout)
{
  command_t& command = *(command_t *)&packet.data;
  //
  // EPD_BEGIN_PAINT
  //
  packet.length = PACKET_LENGTH(command.args.color);
  command.args.color = color;
  command.command = EPD_BEGIN_PAINT;
  if (!device_command(packet, debugout))
    return 1;
  //
  // EPD_IMAGE_DATA
  //
  char *image = bitmap->image();
  int   w     = abs(bitmap->width());
  int   h     = bitmap->height();
  int   line  = bitmap->lineBytes();
  if (h <= 0)
    h = - h;
  else
  {
    image += line * (h - 1);
    line = - line;
  }
  int size  = IMAGE_BYTES(w, h);
  int count = 0;
  while ((count < size) && h--)
  {
    char  *buf = image;
    size_t len = IMAGE_BYTES(w, 1);
    while (len && size)
    {
      int n = size - count;
      if (n > sizeof(command.args.data))
        n = sizeof(command.args.data);
      if (n > len)
        n = len;
      packet.length = PACKET_LENGTH(command.command) + n;
      memcpy(command.args.data, buf, n);
      command.command = EPD_IMAGE_DATA;
      if (!device_command(packet, debugout))
        return 2;
      buf   += n;
      len   -= n;
      count += n;
    }
    image += line;
  }
#if !EPAPER_FAST_IO
  size = IMAGE_BYTES(width, height);;
  while (count < size)
  {
    size_t n = size - count;
    if (n > sizeof(command.args.data))
      n = sizeof(command.args.data);
    packet.length = PACKET_LENGTH(command.command) + n;
    memset(command.args.data, 0, n);
    command.command = EPD_IMAGE_DATA;
    if (!device_command(packet, debugout))
      return 3;
    count += n;
  }
#endif
  //
  // EPD_END_PAINT
  //
  packet.length = PACKET_LENGTH(command.command);
  command.command = EPD_END_PAINT;
  int rv = device_command(packet, debugout) ? 0 : 4;
  if (debugout)
  {
    fprintf(stdout, "\n");
    fflush(stdout);
  }
  return rv;
}

unsigned int read32(const char *buf)
{
  return ((unsigned int)buf[0] << 24)
       | ((unsigned int)buf[1] << 16)
       | ((unsigned int)buf[2] <<  8)
       | ((unsigned int)buf[3] <<  0);
}

void write32(char *buf, unsigned int data)
{
  buf[0] = data >> 24;
  buf[1] = data >> 16;
  buf[2] = data >>  8;
  buf[3] = data >>  0;
}

void device_response(Bridge::packet_t& packet)
{
  int wakeup, err = 0;

  command_t& command = *(command_t *)&packet.data;
  if ((command.command != EPD_QUERY_PAINT) || (command.args.status != 0))
    return;
  //
  // Response of EPD_QUERY_PAINT
  //
  wakeup = calcWakeupTime(read32((char *)&command.args.query.interval), interval, origin);
  writeFile(stdout, LOGFILE, "a", true, "[INFO] query new image (%d, %02d:%02d:%02d)\n"
    , update
    , (wakeup / 3600) % 99
    , (wakeup /   60) % 60
    , (wakeup /    1) % 60
  );
  packet.length = PACKET_LENGTH(command.args.query);
  command.args.status = !update;
  write32((char *)&command.args.query.interval, wakeup);
  if (!device_write(packet))
    return;
  if (!update)
    return;
  //
  // Transfer Bitmap Images
  //
  time_t t = time(NULL);
  for (int color = 0; !err && (color < EPD_COLOR_MAX); ++color)
  {
    Bitmap *bmp = images[color];
    if (bmp)
    {
      writeFile(stdout, LOGFILE, "a", true, "[INFO] image transfer start (color=%d) ", color);
      err = device_paint(packet, color, bmp, DEBUG_RETRY);
      writeFile(nullptr, LOGFILE, "a", false, "\n");
    }
  }
  //
  // EPD_REFRESH
  //
  if (!err)
  {
    packet.length = PACKET_LENGTH(command.command);
    command.command = EPD_REFRESH;
    if (!device_command(packet))
      err = 5;
    else
      update = false;
  }
  t = time(NULL) - t;
  if (err)
    writeFile(stdout, LOGFILE, "a", true, "[ERROR] image transfer failed = %d (%d sec)\n", err, t);
  else
    writeFile(stdout, LOGFILE, "a", true, "[INFO] image transfer success (%d sec)\n", t);
}

void image_check(void)
{
  char backup[256];
  Bitmap bmp;
  bool found = false;

  for (int color = 0; color < EPD_COLOR_MAX; ++color)
  {
    if (plates[color])
    {
      if (bmp.read(plates[color]))
      {
        if ((bmp.bitCount() == bitcount) && (abs(bmp.width()) == width))
        {
          if (images[color] == nullptr)
            images[color] = new Bitmap();
          images[color]->copy(bmp);
          if (!found)
            fprintf(stdout, "[INFO] new image found\n");
          found = true;
        }
        snprintf(backup, sizeof(backup), "%s.backup", plates[color]);
        unlink(backup);
        rename(plates[color], backup);
      }
    }
  }
  if (found)
    update = true;
}

void sigint_handler(int signum)
{
  exitcode = 1;
}

void usage()
{
  fprintf(stdout, "Name Plate Manager, Version 1.0\n");
  fprintf(stdout, "Copyright(c) 2023 Sasapea's Lab. All right reserved.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "Usage: %s [options]\n", prog_name + 1);
  fprintf(stdout, "\n");
  fprintf(stdout, "  options:\n");
  fprintf(stdout, "    -device name ... ftdi_sio device name (ttyUSB0)\n");
  fprintf(stdout, "    -width  n    ... display image width (%d) pixel\n", width);
  fprintf(stdout, "    -height n    ... display image height (%d) pixel\n", height);
  fprintf(stdout, "    -interval n  ... update interval (%d) sec\n", interval);
  fprintf(stdout, "    -origin n    ... update origin (%d) sec\n", origin);
  fprintf(stdout, "    -help        ... this help\n\n");
}

int main(int argc, char **argv)
{
  const char *name = "ttyUSB0"; // ftdi_sio
  bool help        = false;
  struct sigaction nact, oact;
  int i;

  exitcode = 1;
  //
  // save this program path
  //
  prog_path = argv[0];
  prog_name = strrchr(prog_path, '/');
  //
  // setup signal handler
  //
  sigemptyset(&nact.sa_mask);
  nact.sa_handler = sigint_handler;
  nact.sa_flags   = SA_RESTART;
  sigaction(SIGINT , &nact, &oact);
  sigaction(SIGTERM, &nact, &oact);
  //
  // Check Command Line Argument's
  //
  for (i = 1; i < argc; )
  {
    char *p = argv[i++];
    if (*p == '-')
    {
      ++p;
      if (strcmp(p, "debug") == 0)
        debug = true;
      else if (i < argc)
      {
        if (strcmp(p, "device") == 0)
          name = argv[i++];
        else if (strcmp(p, "width") == 0)
          width = atoi(argv[i++]);
        else if (strcmp(p, "height") == 0)
          height = atoi(argv[i++]);
        else if (strcmp(p, "interval") == 0)
          interval = atoi(argv[i++]);
        else if (strcmp(p, "origin") == 0)
          origin = atoi(argv[i++]);
        else
          help = true;
      }
      else
        help = true;
    }
    else
      help = true;
  }
  //
  // Help
  //
  if (help)
    usage();
  //
  // Open Device
  //
  else if (device_open(name, SERIAL_SPEED))
  {
    //
    // Read from Device
    //
    exitcode = 0;
    while (exitcode == 0)
    {
      image_check();
      if ((i = device_read(request, IMAGE_POLLING_TIME)) > 0)
        device_response(request);
      else if (i < 0)
        exitcode = 1;
      fflush(stdout);
    }
    //
    // Close Device
    //
    device_close();
  }
  return exitcode;
}
