/*
 serial.h - Serial I/O Library

 Copyright (c) 2021 Sasapea's Lab. All right reserved.

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General
 Public License along with this library; if not, write to the
 Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 Boston, MA  02111-1307  USA
 */
#ifndef _SERIAL_H
#define _SERIAL_H

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <fcntl.h>

#ifdef __linux__
  #include <unistd.h>
  #include <termios.h>
  #include <sys/time.h>
  #include <sys/types.h>
  #define DEVICE_PATH "/dev/%s"
  #define OPEN_MODE (O_RDWR | O_NOCTTY)
#else
  #include <windows.h>
  #if SIZE_MAX == UINT64_MAX
    #define ssize_t long long
  #else
    #define ssize_t int
  #endif
  #define DEVICE_PATH "\\\\.\\%s"
  #define OPEN_MODE (O_RDWR | _O_BINARY)
#endif


typedef enum
{
  SERIAL_FRAME_5N1,
  SERIAL_FRAME_6N1,
  SERIAL_FRAME_7N1,
  SERIAL_FRAME_8N1,
  SERIAL_FRAME_5N2,
  SERIAL_FRAME_6N2,
  SERIAL_FRAME_7N2,
  SERIAL_FRAME_8N2,
  SERIAL_FRAME_5O1,
  SERIAL_FRAME_6O1,
  SERIAL_FRAME_7O1,
  SERIAL_FRAME_8O1,
  SERIAL_FRAME_5O2,
  SERIAL_FRAME_6O2,
  SERIAL_FRAME_7O2,
  SERIAL_FRAME_8O2,
  SERIAL_FRAME_5E1,
  SERIAL_FRAME_6E1,
  SERIAL_FRAME_7E1,
  SERIAL_FRAME_8E1,
  SERIAL_FRAME_5E2,
  SERIAL_FRAME_6E2,
  SERIAL_FRAME_7E2,
  SERIAL_FRAME_8E2,
} SERIAL_FRAME;

typedef enum
{
  SERIAL_FLOW_NONE,
  SERIAL_FLOW_RTSCTS,
  SERIAL_FLOW_XONOFF,
} SERIAL_FLOW;

#define SERIAL_FLOW_XONOFF_START 0x11 // DC1
#define SERIAL_FLOW_XONOFF_STOP  0x13 // DC3
#define SERIAL_FLOW_XONOFF_LIMIT 256

class Serial
{
  private:

    int  _fd;
    long _timeout;
#ifdef __linux__
    struct termios _tcattr;
#else
    COMMTIMEOUTS _comtimeouts;
#endif

    void
    config(unsigned long speed, SERIAL_FRAME frame, SERIAL_FLOW flow, long timeout)
    {
#ifdef __linux__
      struct termios tcattr;
      tcgetattr(_fd, &tcattr);
      _tcattr= tcattr;
      cfmakeraw(&tcattr);
      speed_t speed0 = cfgetospeed(&tcattr);
      int cflags = tcattr.c_cflag & (CSIZE | CSTOPB | PARENB | PARODD);
      tcattr.c_cflag &= ~(CSIZE | CSTOPB | PARENB | PARODD);
      switch (speed)
      {
        case 50: speed0 = B50; break;
        case 75: speed0 = B75; break;
        case 110: speed0 = B110; break;
        case 134: speed0 = B134; break;
        case 150: speed0 = B150; break;
        case 200: speed0 = B200; break;
        case 300: speed0 = B300; break;
        case 600: speed0 = B600; break;
        case 1200: speed0 = B1200; break;
        case 1800: speed0 = B1800; break;
        case 2400: speed0 = B2400; break;
        case 4800: speed0 = B4800; break;
        case 9600: speed0 = B9600; break;
        case 19200: speed0 = B19200; break;
        case 38400: speed0 = B38400; break;
        case 57600: speed0 = B57600; break;
        case 115200: speed0 = B115200; break;
        case 230400: speed0 = B230400; break;
        case 460800: speed0 = B460800; break;
        case 500000: speed0 = B500000; break;
        case 576000: speed0 = B576000; break;
        case 921600: speed0 = B921600; break;
        case 1000000: speed0 = B1000000; break;
        case 1152000: speed0 = B1152000; break;
        case 1500000: speed0 = B1500000; break;
        case 2000000: speed0 = B2000000; break;
        case 2500000: speed0 = B2500000; break;
        case 3000000: speed0 = B3000000; break;
        case 3500000: speed0 = B3500000; break;
        case 4000000: speed0 = B4000000; break;
      }
      switch (frame)
      {
        case SERIAL_FRAME_5N1: cflags = CS5; break;
        case SERIAL_FRAME_6N1: cflags = CS6; break;
        case SERIAL_FRAME_7N1: cflags = CS7; break;
        case SERIAL_FRAME_8N1: cflags = CS8; break;
        case SERIAL_FRAME_5N2: cflags = CS5 | CSTOPB; break;
        case SERIAL_FRAME_6N2: cflags = CS6 | CSTOPB; break;
        case SERIAL_FRAME_7N2: cflags = CS7 | CSTOPB; break;
        case SERIAL_FRAME_8N2: cflags = CS8 | CSTOPB; break;
        case SERIAL_FRAME_5O1: cflags = CS5 | PARENB | PARODD; break;
        case SERIAL_FRAME_6O1: cflags = CS6 | PARENB | PARODD; break;
        case SERIAL_FRAME_7O1: cflags = CS7 | PARENB | PARODD; break;
        case SERIAL_FRAME_8O1: cflags = CS8 | PARENB | PARODD; break;
        case SERIAL_FRAME_5O2: cflags = CS5 | CSTOPB | PARENB | PARODD; break;
        case SERIAL_FRAME_6O2: cflags = CS6 | CSTOPB | PARENB | PARODD; break;
        case SERIAL_FRAME_7O2: cflags = CS7 | CSTOPB | PARENB | PARODD; break;
        case SERIAL_FRAME_8O2: cflags = CS8 | CSTOPB | PARENB | PARODD; break;
        case SERIAL_FRAME_5E1: cflags = CS5 | PARENB; break;
        case SERIAL_FRAME_6E1: cflags = CS6 | PARENB; break;
        case SERIAL_FRAME_7E1: cflags = CS7 | PARENB; break;
        case SERIAL_FRAME_8E1: cflags = CS8 | PARENB; break;
        case SERIAL_FRAME_5E2: cflags = CS5 | CSTOPB | PARENB; break;
        case SERIAL_FRAME_6E2: cflags = CS6 | CSTOPB | PARENB; break;
        case SERIAL_FRAME_7E2: cflags = CS7 | CSTOPB | PARENB; break;
        case SERIAL_FRAME_8E2: cflags = CS8 | CSTOPB | PARENB; break;
      }
      cfsetspeed(&tcattr, speed0);
      tcattr.c_cflag = (tcattr.c_cflag & ~(CSIZE | CSTOPB | PARENB | PARODD)) | cflags;
#ifdef CRTSCTS
      if (flow == SERIAL_FLOW_RTSCTS)
        tcattr.c_iflag |= CRTSCTS;
      else
        tcattr.c_iflag &= ~CRTSCTS;
#endif
      if (flow == SERIAL_FLOW_XONOFF)
        tcattr.c_iflag |= IXON;
      else
        tcattr.c_iflag &= ~IXON;
      tcattr.c_cc[VSTART] = SERIAL_FLOW_XONOFF_START;
      tcattr.c_cc[VSTOP] = SERIAL_FLOW_XONOFF_STOP;
      tcattr.c_oflag &= ~OPOST;
      tcflush(_fd, TCIOFLUSH);
      tcsetattr(_fd, TCSANOW, &tcattr);
#else
      DCB dcb;
      HANDLE hSerial = (HANDLE) _get_osfhandle(_fd);
      memset(&dcb, 0, sizeof(DCB));
      dcb.DCBlength = sizeof(DCB);
      GetCommState(hSerial, &dcb);
      dcb.BaudRate = speed;
      dcb.ByteSize = (frame & 3) + 5;
      dcb.StopBits = (frame >> 1) & 2;
      dcb.Parity = (frame >> 3);
      dcb.fDtrControl = (flow == SERIAL_FLOW_RTSCTS
        ? DTR_CONTROL_HANDSHAKE : DTR_CONTROL_DISABLE);
      dcb.fRtsControl = (flow == SERIAL_FLOW_RTSCTS
        ? RTS_CONTROL_HANDSHAKE : DTR_CONTROL_DISABLE);
      dcb.fOutxCtsFlow = (flow == SERIAL_FLOW_RTSCTS);
      dcb.fOutxDsrFlow = (flow == SERIAL_FLOW_RTSCTS);
      dcb.fOutX = (flow == SERIAL_FLOW_XONOFF);
      dcb.fInX = (flow == SERIAL_FLOW_XONOFF);
      dcb.XonLim = SERIAL_FLOW_XONOFF_LIMIT;
      dcb.XoffLim = SERIAL_FLOW_XONOFF_LIMIT;
      dcb.XonChar = SERIAL_FLOW_XONOFF_START;
      dcb.XoffChar = SERIAL_FLOW_XONOFF_STOP;
      dcb.fTXContinueOnXoff = TRUE;
      dcb.fDsrSensitivity = FALSE;
      dcb.fErrorChar = FALSE;
      dcb.fNull = FALSE;
      dcb.fAbortOnError = FALSE;
      dcb.fBinary = TRUE;
      SetCommState(hSerial, &dcb);
      PurgeComm(hSerial,
        PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);
      memset(&_comtimeouts, 0, sizeof(_comtimeouts));
      _comtimeouts.ReadIntervalTimeout        = MAXDWORD;
      _comtimeouts.ReadTotalTimeoutMultiplier = MAXDWORD;
      _comtimeouts.ReadTotalTimeoutConstant   = timeout;
      _comtimeouts.WriteTotalTimeoutConstant  = timeout;
      SetCommTimeouts(hSerial, &_comtimeouts);
#endif
      _timeout = timeout;
    }

  public:

    Serial(void)
    : _fd(-1)
    {
    }

    virtual
    ~Serial(void)
    {
      close();
    }

    virtual int
    open(const char *device,
         unsigned long speed = 115200,
         SERIAL_FRAME frame = SERIAL_FRAME_8N1,
         SERIAL_FLOW flow = SERIAL_FLOW_NONE,
         unsigned long timeout = 1000)
    {
      char path[32];
      snprintf(path, sizeof(path), DEVICE_PATH, device);
      if ((_fd = ::open(path, OPEN_MODE)) >= 0)
      {
#ifdef __linux__
        if (lockf(_fd, F_TLOCK, 0) < 0)
        {
          ::close(_fd);
          return _fd = -1;
        }
#endif
        config(speed, frame, flow, timeout);
      }
      return _fd;
    }

    virtual int
    close(void)
    {
      int rv = 0;
      if (_fd >= 0)
      {
#ifdef __linux__
        tcsetattr(_fd, TCSANOW, &_tcattr);
#endif
        rv = ::close(_fd);
        _fd = -1;
      }
      return rv;
    }

    virtual ssize_t
    read(void *buf, size_t len, long timeout = -1)
    {
      int rv;
      if (timeout < 0)
        timeout = _timeout;
#ifdef __linux__
      struct timeval tov = { timeout / 1000, (timeout % 1000) * 1000 };
      fd_set rdfds;
      FD_ZERO(&rdfds);
      FD_SET(_fd, &rdfds);
      if ((rv = select(_fd + 1, &rdfds, NULL, NULL, &tov)) > 0)
        rv = ::read(_fd, buf, len);
#else
      if (_comtimeouts.ReadTotalTimeoutConstant != timeout)
      {
        _comtimeouts.ReadTotalTimeoutConstant = timeout;
        SetCommTimeouts((HANDLE)_get_osfhandle(_fd), &_comtimeouts);
      }
      rv = ::read(_fd, buf, len);
#endif
      return rv;
    }

    virtual ssize_t
    write(const void *buf, size_t len, long timeout = -1)
    {
      int rv;
      if (timeout < 0)
        timeout = _timeout;
#ifdef __linux__
      struct timeval tov = { timeout / 1000, (timeout % 1000) * 1000 };
      fd_set wrfds;
      FD_ZERO(&wrfds);
      FD_SET(_fd, &wrfds);
      if ((rv = select(_fd + 1, NULL, &wrfds,  NULL, &tov)) > 0)
        rv = ::write(_fd, buf, len);
#else
      if (_comtimeouts.WriteTotalTimeoutConstant != timeout)
      {
        _comtimeouts.WriteTotalTimeoutConstant = timeout;
        SetCommTimeouts((HANDLE)_get_osfhandle(_fd), &_comtimeouts);
      }
      rv = ::write(_fd, buf, len);
#endif
      return rv;
    }
};

#endif
